'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _ramda;

function _load_ramda() {
  return _ramda = _interopRequireDefault(require('ramda'));
}

var _chalk;

function _load_chalk() {
  return _chalk = _interopRequireDefault(require('chalk'));
}

var _objectDiff;

function _load_objectDiff() {
  return _objectDiff = _interopRequireDefault(require('object-diff'));
}

var _ttyTable;

function _load_ttyTable() {
  return _ttyTable = _interopRequireDefault(require('tty-table'));
}

var _normalizePath;

function _load_normalizePath() {
  return _normalizePath = _interopRequireDefault(require('normalize-path'));
}

var _arrayDifference;

function _load_arrayDifference() {
  return _arrayDifference = _interopRequireDefault(require('array-difference'));
}

var _consumerComponent;

function _load_consumerComponent() {
  return _consumerComponent = _interopRequireDefault(require('../../consumer/component/consumer-component'));
}

var _docsTemplate;

function _load_docsTemplate() {
  return _docsTemplate = _interopRequireDefault(require('./docs-template'));
}

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var fields = ['id', 'compiler', 'tester', 'language', 'mainFile', 'dependencies', 'devDependencies', 'packages', 'devPackages', 'files', 'specs', 'deprecated'];

var header = [{ value: 'ID', width: 20, headerColor: 'cyan', headerAlign: 'left', align: 'left' }];
var opts = {
  borderStyle: 1,
  paddingBottom: 0,
  paddingLeft: 0,
  paddingRight: 1,
  align: 'center',
  color: 'white'
};
function comparator(a, b) {
  if (a instanceof Array && b instanceof Array) {
    return (_ramda || _load_ramda()).default.isEmpty((0, (_arrayDifference || _load_arrayDifference()).default)(a, b));
  }
  return a === b;
}

function convertObjectToPrintable(component, isFromFs) {
  var obj = {};
  var parsePackages = function parsePackages(packages) {
    return !(_ramda || _load_ramda()).default.isEmpty(packages) && !(_ramda || _load_ramda()).default.isNil(packages) ? Object.keys(packages).map(function (key) {
      return `${key}@${packages[key]}`;
    }) : null;
  };
  var name = component.name,
      box = component.box,
      lang = component.lang,
      compilerId = component.compilerId,
      testerId = component.testerId,
      dependencies = component.dependencies,
      devDependencies = component.devDependencies,
      packageDependencies = component.packageDependencies,
      devPackageDependencies = component.devPackageDependencies,
      files = component.files,
      mainFile = component.mainFile,
      deprecated = component.deprecated,
      version = component.version,
      docs = component.docs;

  obj.id = isFromFs ? `${box}/${name}@${version} [file system]` : `${box}/${name}@${version}`;
  obj.compiler = compilerId ? compilerId.toString() : null;
  obj.language = lang || null;
  obj.tester = testerId ? testerId.toString() : null;
  obj.mainFile = mainFile ? (0, (_normalizePath || _load_normalizePath()).default)(mainFile) : null;
  obj.dependencies = dependencies.toStringOfIds().concat(parsePackages(packageDependencies));
  obj.devDependencies = devDependencies.toStringOfIds().concat(parsePackages(devPackageDependencies));

  obj.files = !(_ramda || _load_ramda()).default.isEmpty(files) && !(_ramda || _load_ramda()).default.isNil(files) ? files.filter(function (file) {
    return !file.test;
  }).map(function (file) {
    return (0, (_normalizePath || _load_normalizePath()).default)(file.relative);
  }) : null;
  obj.specs = !(_ramda || _load_ramda()).default.isEmpty(files) && !(_ramda || _load_ramda()).default.isNil(files) && (_ramda || _load_ramda()).default.find((_ramda || _load_ramda()).default.propEq('test', true))(files) ? files.filter(function (file) {
    return file.test;
  }).map(function (file) {
    return (0, (_normalizePath || _load_normalizePath()).default)(file.relative);
  }) : null;
  obj.deprecated = deprecated ? 'True' : null;
  obj.docs = docs;
  return obj;
}

function generateDependenciesTable(component, showRemoteVersion) {
  if (!component.hasDependencies()) {
    return '';
  }
  var dependencyHeader = [];
  if (showRemoteVersion) {
    dependencyHeader.push({
      value: 'Dependency ID',
      width: 50,
      headerColor: 'cyan',
      headerAlign: 'left',
      align: 'left'
    });
    dependencyHeader.push({
      value: 'Current Version',
      width: 20,
      headerColor: 'cyan',
      headerAlign: 'left',
      align: 'left'
    });
    dependencyHeader.push({
      value: 'Local Version',
      width: 20,
      headerColor: 'cyan',
      headerAlign: 'left',
      align: 'left'
    });
    dependencyHeader.push({
      value: 'Remote Version',
      width: 20,
      headerColor: 'cyan',
      headerAlign: 'left',
      align: 'left'
    });
  } else {
    dependencyHeader.push({ value: 'Dependencies', width: 50, headerColor: 'cyan' });
  }
  var getDependenciesRows = function getDependenciesRows(dependencies) {
    var isDev = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

    var dependencyRows = [];
    dependencies.forEach(function (dependency) {
      var dependencyId = showRemoteVersion ? dependency.id.toStringWithoutVersion() : dependency.id.toString();
      dependencyId = isDev ? `${dependencyId} (dev)` : dependencyId;
      var row = [dependencyId];
      if (showRemoteVersion) {
        var dependencyVersion = parseInt(dependency.currentVersion);
        var localVersion = parseInt(dependency.localVersion);
        var remoteVersion = dependency.remoteVersion ? parseInt(dependency.remoteVersion) : null;
        // if all versions are equal, paint them with green. Otherwise, paint with red
        var color = remoteVersion && remoteVersion === localVersion && remoteVersion === dependencyVersion || !remoteVersion && localVersion === dependencyVersion ? 'green' : 'red';
        row.push((_chalk || _load_chalk()).default[color](dependencyVersion));
        row.push((_chalk || _load_chalk()).default[color](localVersion));
        row.push(remoteVersion ? (_chalk || _load_chalk()).default[color](remoteVersion) : 'N/A');
      }
      dependencyRows.push(row);
    });
    return dependencyRows;
  };
  var dependenciesRows = getDependenciesRows(component.dependencies.get());
  var devDependenciesRows = getDependenciesRows(component.devDependencies.get(), true);
  var allDependenciesRows = (_ramda || _load_ramda()).default.concat(dependenciesRows, devDependenciesRows);

  var dependenciesTable = new (_ttyTable || _load_ttyTable()).default(dependencyHeader, allDependenciesRows);
  return dependenciesTable.render();
}

function paintWithoutCompare(component, showRemoteVersion) {
  var printableComponent = convertObjectToPrintable(component, false);
  var rows = fields.map(function (field) {
    var title = `${field[0].toUpperCase()}${field.substr(1)}`.replace(/([A-Z])/g, ' $1').trim();
    var arr = [];
    if (!printableComponent[field]) return null;
    if (field === 'id') {
      header.push({
        value: printableComponent[field],
        width: 50,
        headerColor: 'white',
        headerAlign: 'left',
        align: 'left'
      });
      return null;
    }
    arr.push((_chalk || _load_chalk()).default.cyan(title));
    if (!printableComponent[field]) return null;

    if (printableComponent[field]) {
      if (printableComponent[field] instanceof Array) {
        arr.push(printableComponent[field].join('\n'));
      } else arr.push(printableComponent[field]);
    }
    return arr;
  }).filter(function (x) {
    return x;
  });

  var componentTable = new (_ttyTable || _load_ttyTable()).default(header, rows, opts);
  var componentTableStr = componentTable.render();
  var dependenciesTableStr = showRemoteVersion ? generateDependenciesTable(component, showRemoteVersion) : '';

  return componentTableStr + dependenciesTableStr + (0, (_docsTemplate || _load_docsTemplate()).default)(component.docs);
}

function paintWithCompare(originalComponent, componentToCompareTo, showRemoteVersion) {
  var printableOriginalComponent = convertObjectToPrintable(originalComponent, true);
  var printableComponentToCompare = convertObjectToPrintable(componentToCompareTo, false);

  var componentsDiffs = (_objectDiff || _load_objectDiff()).default.custom({
    equal: comparator
  }, printableOriginalComponent, printableComponentToCompare);

  var rows = fields.map(function (field) {
    if (!printableOriginalComponent[field] && !printableComponentToCompare[field]) return null;
    var title = `${field[0].toUpperCase()}${field.substr(1)}`.replace(/([A-Z])/g, ' $1').trim();
    if (field === 'id') {
      header.push({
        value: printableComponentToCompare[field],
        width: 50,
        headerColor: 'white',
        headerAlign: 'left',
        align: 'left'
      });
      header.push({
        value: printableOriginalComponent[field],
        width: 50,
        headerColor: 'white',
        headerAlign: 'left',
        align: 'left'
      });
      return null;
    }
    var arr = field in componentsDiffs && field !== 'id' ? [(_chalk || _load_chalk()).default.red(title)] : [(_chalk || _load_chalk()).default.cyan(title)];
    printableComponentToCompare[field] instanceof Array ? arr.push(printableComponentToCompare[field].join('\n')) : arr.push(printableComponentToCompare[field]);
    printableOriginalComponent[field] instanceof Array ? arr.push(printableOriginalComponent[field].join('\n')) : arr.push(printableOriginalComponent[field]);
    return arr;
  }).filter(function (x) {
    return x;
  });

  var componentTable = new (_ttyTable || _load_ttyTable()).default(header, rows);
  var componentTableStr = componentTable.render();
  var dependenciesTableStr = !componentToCompareTo ? generateDependenciesTable(originalComponent, showRemoteVersion) : '';
  return componentTableStr + dependenciesTableStr;
}

exports.default = function (component, componentModel, showRemoteVersion) {
  return componentModel ? paintWithCompare(component, componentModel, showRemoteVersion) : paintWithoutCompare(component, showRemoteVersion);
};
//# sourceMappingURL=component-template.js.map