'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = linkComponents;

var _path;

function _load_path() {
  return _path = _interopRequireDefault(require('path'));
}

var _fsExtra;

function _load_fsExtra() {
  return _fsExtra = _interopRequireDefault(require('fs-extra'));
}

var _ramda;

function _load_ramda() {
  return _ramda = _interopRequireDefault(require('ramda'));
}

var _symlinkOrCopy;

function _load_symlinkOrCopy() {
  return _symlinkOrCopy = _interopRequireDefault(require('symlink-or-copy'));
}

var _glob;

function _load_glob() {
  return _glob = _interopRequireDefault(require('glob'));
}

var _bitId;

function _load_bitId() {
  return _bitId = require('../bit-id');
}

var _component;

function _load_component() {
  return _component = _interopRequireDefault(require('../consumer/component'));
}

var _constants;

function _load_constants() {
  return _constants = require('../constants');
}

var _componentMap;

function _load_componentMap() {
  return _componentMap = _interopRequireDefault(require('../consumer/bit-map/component-map'));
}

var _logger;

function _load_logger() {
  return _logger = _interopRequireDefault(require('../logger/logger'));
}

var _utils;

function _load_utils() {
  return _utils = require('../utils');
}

var _consumer;

function _load_consumer() {
  return _consumer = _interopRequireDefault(require('../consumer/consumer'));
}

var _linkGenerator;

function _load_linkGenerator() {
  return _linkGenerator = require('./link-generator');
}

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

/**
 * @param componentId
 * @param srcPath the path where the symlink is pointing to
 * @param destPath the path where to write the symlink
 */
function createSymlinkOrCopy(componentId, srcPath, destPath) {
  (_fsExtra || _load_fsExtra()).default.removeSync(destPath); // in case a component has been moved
  (_fsExtra || _load_fsExtra()).default.ensureDirSync((_path || _load_path()).default.dirname(destPath));
  try {
    (_logger || _load_logger()).default.debug(`generating a symlink on ${destPath} pointing to ${srcPath}`);
    (_symlinkOrCopy || _load_symlinkOrCopy()).default.sync(srcPath, destPath);
  } catch (err) {
    throw new Error(`failed to link a component ${componentId.toString()}.
         Symlink (or maybe copy for Windows) from: ${srcPath}, to: ${destPath} was failed.
         Original error: ${err}`);
  }
}

function writeDependencyLink(parentRootDir, bitId, rootDir, bindingPrefix) {
  var relativeDestPath = (_consumer || _load_consumer()).default.getNodeModulesPathOfComponent(bindingPrefix, bitId);
  var destPath = (_path || _load_path()).default.join(parentRootDir, relativeDestPath);
  createSymlinkOrCopy(bitId, rootDir, destPath);

  return { from: parentRootDir, to: rootDir };
}

/**
 * When the dists is outside the components directory, it doesn't have access to the node_modules of the component's
 * root-dir. The solution is to go through the node_modules packages one by one and symlink them.
 */
function symlinkPackages(from, to, consumer) {
  var dependenciesSavedAsComponents = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : true;

  var fromNodeModules = (_path || _load_path()).default.join(from, 'node_modules');
  var toNodeModules = (_path || _load_path()).default.join(to, 'node_modules');
  var dirsIncludesBindingPrefix = (_glob || _load_glob()).default.sync('*', { cwd: fromNodeModules });
  // when dependenciesSavedAsComponents the node_modules/@bit has real link files, we don't want to touch them
  // otherwise, node_modules/@bit has packages as any other directory in node_modules
  var dirs = dependenciesSavedAsComponents ? dirsIncludesBindingPrefix.filter(function (dir) {
    return dir !== consumer.bitJson.bindingPrefix;
  }) : dirsIncludesBindingPrefix;
  if (!dirs.length) return;
  (_logger || _load_logger()).default.debug(`deleting the content of ${toNodeModules}`);
  (_fsExtra || _load_fsExtra()).default.emptyDirSync(toNodeModules);
  dirs.forEach(function (dir) {
    (_symlinkOrCopy || _load_symlinkOrCopy()).default.sync((_path || _load_path()).default.join(fromNodeModules, dir), (_path || _load_path()).default.join(toNodeModules, dir));
  });
}

function writeDependenciesLinks(component, componentMap, consumer) {
  return component.getAllDependencies().map(function (dependency) {
    var dependencyComponentMap = consumer.bitMap.getComponent(dependency.id);
    var writtenLinks = [];
    if (!dependencyComponentMap) return writtenLinks;
    writtenLinks.push(writeDependencyLink((_path || _load_path()).default.join(consumer.getPath(), componentMap.rootDir), dependency.id, (_path || _load_path()).default.join(consumer.getPath(), dependencyComponentMap.rootDir), component.bindingPrefix));
    if (!consumer.shouldDistsBeInsideTheComponent()) {
      var _from = component.dists.getDistDirForConsumer(consumer, componentMap.rootDir);
      var _to = component.dists.getDistDirForConsumer(consumer, dependencyComponentMap.rootDir);
      writtenLinks.push(writeDependencyLink(_from, dependency.id, _to, component.bindingPrefix));
      symlinkPackages(_from, _to, consumer);
    }
    return writtenLinks;
  });
}

/**
 * Link from node_modules/@bit/component-name/index.js to the component's main file.
 * It is needed for Authored components only.
 * Since an authored component doesn't have rootDir, it's impossible to symlink to the component directory.
 * It makes it easier for Author to use absolute syntax between their own components.
 */
function linkToMainFile(component, componentMap, componentId, consumer) {
  component.dists.updateDistsPerConsumerBitJson(component.id, consumer, componentMap);
  var mainFile = component.dists.calculateMainDistFileForAuthored(component.mainFile, consumer);
  var indexFileName = (0, (_linkGenerator || _load_linkGenerator()).getIndexFileName)(mainFile);
  var dest = (_path || _load_path()).default.join((_consumer || _load_consumer()).default.getNodeModulesPathOfComponent(component.bindingPrefix, componentId), indexFileName);
  var destRelative = (0, (_utils || _load_utils()).pathRelativeLinux)((_path || _load_path()).default.dirname(dest), mainFile);
  var fileContent = (0, (_linkGenerator || _load_linkGenerator()).getLinkContent)(destRelative);
  (_fsExtra || _load_fsExtra()).default.outputFileSync(dest, fileContent);
}

function writeMissingLinks(component, componentMap, bitMap) {
  return component.missingDependencies.missingLinks.map(function (dependencyIdStr) {
    var dependencyId = bitMap.getExistingComponentId(dependencyIdStr);
    if (!dependencyId) return null;

    var dependencyComponentMap = bitMap.getComponent(dependencyId);
    return writeDependencyLink(componentMap.rootDir, (_bitId || _load_bitId()).BitId.parse(dependencyId), dependencyComponentMap.rootDir, component.bindingPrefix);
  });
}

function linkComponents(components, consumer) {
  return components.map(function (component) {
    var componentId = component.id;
    (_logger || _load_logger()).default.debug(`linking component to node_modules: ${componentId}`);
    var componentMap = consumer.bitMap.getComponent(componentId, true);
    if (componentMap.origin === (_constants || _load_constants()).COMPONENT_ORIGINS.IMPORTED) {
      var relativeLinkPath = (_consumer || _load_consumer()).default.getNodeModulesPathOfComponent(consumer.bitJson.bindingPrefix, componentId);
      var linkPath = (_path || _load_path()).default.join(consumer.getPath(), relativeLinkPath);
      // when a user moves the component directory, use component.writtenPath to find the correct target
      var srcTarget = component.writtenPath || (_path || _load_path()).default.join(consumer.getPath(), componentMap.rootDir);
      if (!component.dists.isEmpty() && component.dists.writeDistsFiles && !consumer.shouldDistsBeInsideTheComponent()) {
        var distTarget = component.dists.getDistDirForConsumer(consumer, componentMap.rootDir);
        symlinkPackages(srcTarget, distTarget, consumer, component.dependenciesSavedAsComponents);
        createSymlinkOrCopy(componentId, distTarget, linkPath);
      } else {
        createSymlinkOrCopy(componentId, srcTarget, linkPath);
      }

      var _bound = [{ from: componentMap.rootDir, to: relativeLinkPath }];
      var boundDependencies = component.hasDependencies() ? writeDependenciesLinks(component, componentMap, consumer) : [];
      var boundMissingDependencies = component.missingDependencies && component.missingDependencies.missingLinks ? writeMissingLinks(component, componentMap, consumer.bitMap) : [];

      return { id: componentId, bound: _bound.concat([].concat(_toConsumableArray((_ramda || _load_ramda()).default.flatten(boundDependencies)), _toConsumableArray(boundMissingDependencies))) };
    }
    if (componentMap.origin === (_constants || _load_constants()).COMPONENT_ORIGINS.NESTED) {
      if (!component.hasDependencies()) return { id: componentId, bound: [] };
      var _bound2 = writeDependenciesLinks(component, componentMap, consumer);
      return { id: componentId, bound: _bound2 };
    }

    // origin is AUTHORED
    if (!componentId.scope) return { id: componentId, bound: [] }; // scope is a must to generate the link
    var filesToBind = componentMap.getFilesRelativeToConsumer();
    var bound = filesToBind.map(function (file) {
      var dest = (_path || _load_path()).default.join((_consumer || _load_consumer()).default.getNodeModulesPathOfComponent(component.bindingPrefix, componentId), file);
      var destRelative = (0, (_utils || _load_utils()).pathRelativeLinux)((_path || _load_path()).default.dirname(dest), file);
      var fileContent = `module.exports = require('${destRelative}');`;
      (_fsExtra || _load_fsExtra()).default.outputFileSync(dest, fileContent);
      return { from: dest, to: file };
    });
    linkToMainFile(component, componentMap, componentId, consumer);
    return { id: componentId, bound };
  });
}
//# sourceMappingURL=node-modules-linker.js.map