'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.createExtensionLogger = exports.baseFileTransportOpts = undefined;

var _winston;

function _load_winston() {
  return _winston = _interopRequireDefault(require('winston'));
}

var _path;

function _load_path() {
  return _path = _interopRequireDefault(require('path'));
}

var _constants;

function _load_constants() {
  return _constants = require('../constants');
}

var _timers;

function _load_timers() {
  return _timers = require('timers');
}

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Store the extensionsLoggers to prevent create more than one logger for the same extension
// in case the extension developer use api.logger more than once
var extensionsLoggers = new Map();

var baseFileTransportOpts = exports.baseFileTransportOpts = {
  filename: (_path || _load_path()).default.join((_constants || _load_constants()).GLOBAL_LOGS, 'debug.log'),
  json: false,
  // Make it debug level also in production until the product will be more stable. in the future this should be changed to error
  level: process.env.NODE_ENV === 'production' ? 'debug' : 'debug',
  maxsize: 10 * 1024 * 1024, // 10MB
  maxFiles: 10,
  colorize: true,
  prettyPrint: true,
  // If true, log files will be rolled based on maxsize and maxfiles, but in ascending order.
  // The filename will always have the most recent log lines. The larger the appended number, the older the log file
  tailable: true
};

var exceptionsFileTransportOpts = Object.assign({}, baseFileTransportOpts, {
  filename: (_path || _load_path()).default.join((_constants || _load_constants()).GLOBAL_LOGS, 'exceptions.log')
});

var logger = new (_winston || _load_winston()).default.Logger({
  transports: [new (_winston || _load_winston()).default.transports.File(baseFileTransportOpts)],
  exceptionHandlers: [new (_winston || _load_winston()).default.transports.File(exceptionsFileTransportOpts)],
  exitOnError: false
});

/**
 * Create a logger instance for extension
 * The extension name will be added as label so it will appear in the begining of each log line
 * The logger is cached for each extension so there is no problem to use getLogger few times for the same extension
 * @param {string} extensionName
 */
var createExtensionLogger = exports.createExtensionLogger = function createExtensionLogger(extensionName) {
  // Getting logger from cache
  var existingLogger = extensionsLoggers.get(extensionName);

  if (existingLogger) {
    return existingLogger;
  }
  var extensionFileTransportOpts = Object.assign({}, baseFileTransportOpts, {
    filename: (_path || _load_path()).default.join((_constants || _load_constants()).GLOBAL_LOGS, 'extensions.log'),
    label: extensionName
  });
  var extLogger = new (_winston || _load_winston()).default.Logger({
    transports: [new (_winston || _load_winston()).default.transports.File(extensionFileTransportOpts)],
    exceptionHandlers: [new (_winston || _load_winston()).default.transports.File(extensionFileTransportOpts)],
    exitOnError: false
  });
  extensionsLoggers.set(extensionName, extLogger);
  return extLogger;
};

// @credit Kegsay from https://github.com/winstonjs/winston/issues/228
// it solves an issue when exiting the code explicitly and the log file is not written
logger.exitAfterFlush = function () {
  var code = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 0;
  var commandName = arguments[1];

  var level = void 0;
  var msg = void 0;
  if (code === 0) {
    level = 'info';
    msg = `[*] the command ${commandName} has been completed successfully`;
  } else {
    level = 'error';
    msg = `[*] the command ${commandName} has been terminated with an error code ${code}`;
  }
  logger.log(level, msg, function () {
    var numFlushes = 0;
    var numFlushed = 0;
    Object.keys(logger.transports).forEach(function (k) {
      if (logger.transports[k]._stream) {
        numFlushes += 1;
        logger.transports[k]._stream.once('finish', function () {
          numFlushed += 1;
          if (numFlushes === numFlushed) {
            process.exit(code);
          }
        });
        logger.transports[k]._stream.end();
      }
    });
    if (numFlushes === 0) {
      process.exit(code);
    }
  });
};

exports.default = logger;